"""
Utility functions to handle the PARI stack and copy objects from it
*******************************************************************

NOTE: nothing in this module is accessible from Python (only Cython),
so it is not included in the documentation.
"""

#*****************************************************************************
#       Copyright (C) 2016 Luca De Feo <luca.defeo@polytechnique.edu>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 2 of the License, or
# (at your option) any later version.
#                  http://www.gnu.org/licenses/
#*****************************************************************************

from __future__ import absolute_import, division, print_function

cdef extern from *:
    int sig_on_count "cysigs.sig_on_count"
from cysignals.signals cimport sig_off
from cysignals.memory cimport check_malloc

from .paridecl cimport pari_mainstack, avma, paristack_setsize, gsizebyte, gcopy_avma, gnil


cdef inline void clear_stack():
    """
    Call ``sig_off()``. If we are leaving the outermost
    ``sig_on() ... sig_off()`` block, then clear the PARI stack.
    """
    global avma
    if sig_on_count <= 1:
        avma = pari_mainstack.top
    sig_off()

cdef inline GEN deepcopy_to_python_heap(GEN x, void** address) except NULL:
    cdef size_t s = gsizebyte(x)
    cdef void* tmp_bot = check_malloc(s)
    address[0] = tmp_bot
    cdef pari_sp tmp_top = <pari_sp>tmp_bot + s
    return gcopy_avma(x, &tmp_top)

cdef inline Gen new_gen(GEN x):
    """
    Create a new Gen wrapping `x`, then call ``clear_stack()``.
    Except if `x` is ``gnil``, then we return ``None`` instead.
    """
    cdef Gen g
    if x is gnil:
        g = None
    else:
        g = new_gen_noclear(x)
    clear_stack()
    return g

cdef inline Gen new_gen_noclear(GEN x):
    """
    Create a new gen, but don't free any memory on the stack and don't
    call sig_off().
    """
    cdef Gen y = Gen.__new__(Gen)
    y.g = deepcopy_to_python_heap(x, &y.chunk)
    return y



#!/usr/bin/env python

import os
import sys

if "READTHEDOCS" in os.environ:
    # When building with readthedocs, install the dependencies too.
    # See https://github.com/rtfd/readthedocs.org/issues/2776
    for reqs in ["requirements.txt"]:
        if os.path.isfile(reqs):
            from subprocess import check_call
            check_call([sys.executable, "-m", "pip", "install", "-r", reqs])

    # Print PARI/GP defaults and environment variables for debugging
    from subprocess import Popen, PIPE
    Popen(["gp", "-f", "-q"], stdin=PIPE).communicate("default()")
    for item in os.environ.items():
        print("%s=%r" % item)


from setuptools import setup
from distutils.command.build_ext import build_ext as _build_ext
from setuptools.command.bdist_egg import bdist_egg as _bdist_egg
from setuptools.extension import Extension

from autogen import rebuild
from autogen.paths import include_dirs, library_dirs


# Adapted from Cython's new_build_ext
class build_ext(_build_ext):
    def finalize_options(self):
        # Check dependencies
        try:
            from Cython.Build.Dependencies import cythonize
        except ImportError as E:
            sys.stderr.write("Error: {0}\n".format(E))
            sys.stderr.write("The installation of cypari2 requires Cython\n")
            sys.exit(1)

        try:
            # We need the header files for cysignals at compile-time
            import cysignals
        except ImportError as E:
            sys.stderr.write("Error: {0}\n".format(E))
            sys.stderr.write("The installation of cypari2 requires cysignals\n")
            sys.exit(1)

        # Generate auto-generated sources from pari.desc
        rebuild()

        self.directives = {
            "autotestdict.cdef": True,
            "binding": True,
            "cdivision": True,
            "language_level": 2,
        }

        self.distribution.ext_modules[:] = cythonize(
            self.distribution.ext_modules,
            compiler_directives=self.directives)

        _build_ext.finalize_options(self)


class no_egg(_bdist_egg):
    def run(self):
        from distutils.errors import DistutilsOptionError
        raise DistutilsOptionError("The package cypari2 will not function correctly when built as egg. Therefore, it cannot be installed using 'python setup.py install' or 'easy_install'. Instead, use 'pip install' to install cypari2.")


setup(
    name='cypari2',
    version=open("VERSION").read().strip(),
    description='An interface to the number theory library libpari',
    url='https://github.com/defeo/cypari2',
    author='Many people',
    author_email="sage-devel@googlegroups.com",
    license='GNU General Public License, version 2 or later',
    ext_modules=[Extension("*", ["cypari2/*.pyx"],
            include_dirs=include_dirs(), library_dirs=library_dirs())],
    keywords='PARI/GP number theory',
    packages=['cypari2'],
    package_dir={'cypari2': 'cypari2'},
    package_data={'cypari2': ['declinl.pxi', '*.pxd', '*.h']},
    cmdclass=dict(build_ext=build_ext, bdist_egg=no_egg)
)



from .types cimport *
cimport cython


cdef class Gen_auto:
    # The actual PARI GEN
    cdef GEN g

    # Chunk of memory containing g, typically allocated in
    # deepcopy_to_python_heap().
    cdef void* chunk

    # Parent Gen: this is usually None, but it can be used when this
    # Gen is a part of some other Gen. In that case, chunk will be NULL
    # because the parent has allocated the memory.
    cdef Gen_auto parent

    # A cache for __getitem__. Initially, this is None but it will be
    # turned into a dict when needed. This is not just an optional
    # cache, but it is required to implement __setitem__ correctly:
    # __setitem__ will set an entry of one Gen to another Gen. To
    # prevent Python from garbage collecting the other Gen, we must put
    # it in itemcache.
    cdef dict itemcache

    cdef inline int cache(self, key, value) except -1:
        """
        Add ``(key, value)`` to ``self.itemcache``.
        """
        if self.itemcache is None:
            self.itemcache = {}
        self.itemcache[key] = value

@cython.final
cdef class Gen(Gen_auto):
    cdef new_ref(self, GEN g)

cdef Gen list_of_Gens_to_Gen(list s)
cpdef Gen objtogen(s)
